#include "runtime.h"
#include "threading.h"

#ifdef __linux__

    #include <pthread.h>
struct vy_condition_var_s {
    pthread_mutex_t mutex;
    pthread_cond_t cond;
    ptrdiff_t next_reusable;
};

    #define MAX_CONDS 1024
vy_condition_var _conds[MAX_CONDS];
static ptrdiff_t _first_reusable = MAX_CONDS;
static ptrdiff_t _next           = 0;
static pthread_mutex_t _guard    = PTHREAD_MUTEX_INITIALIZER;

VY_DLLEXPORT vy_condition_var *vyCreateConditionVar(void) {
    pthread_mutex_lock(&_guard);
    if (_first_reusable < MAX_CONDS) {
        vy_condition_var *cond = &_conds[_first_reusable];
        _first_reusable        = cond->next_reusable;
        pthread_mutex_unlock(&_guard);
        return cond;
    } else if (_next < MAX_CONDS) {
        vy_condition_var *cond = &_conds[_next];
        if (pthread_mutex_init(&cond->mutex, NULL) != 0) {
            vyLog("core", "Condition variable creation failed");
            pthread_mutex_unlock(&_guard);
            return NULL;
        }
        if (pthread_cond_init(&cond->cond, NULL) != 0) {
            vyLog("core", "Condition variable creation failed");
        }
        cond->next_reusable = MAX_CONDS;
        ++_next;
        pthread_mutex_unlock(&_guard);
        return cond;
    }
    vyReportError("core", "Ran out of condition variable objects");
    pthread_mutex_unlock(&_guard);
    return NULL;
}

VY_DLLEXPORT void vyDestroyConditionVar(vy_condition_var *var) {
    ptrdiff_t index = var - &_conds[0];
    pthread_mutex_lock(&_guard);
    var->next_reusable = _first_reusable;
    _first_reusable    = index;
    pthread_mutex_unlock(&_guard);
}

VY_DLLEXPORT void vyLockConditionVar(vy_condition_var *var) {
    pthread_mutex_lock(&var->mutex);
}

VY_DLLEXPORT void vyUnlockConditionVar(vy_condition_var *var, bool signal) {
    if (signal)
        pthread_cond_signal(&var->cond);
    pthread_mutex_unlock(&var->mutex);
}

VY_DLLEXPORT void vyWaitOnConditionVar(vy_condition_var *var) {
    pthread_cond_wait(&var->cond, &var->mutex);
}

#endif
