#include "runtime.h"
#include "threading.h"

#if defined(_WIN32)
#elif defined(__linux__)

    #include <pthread.h>

struct vy_thread_s {
    pthread_t handle;
    ptrdiff_t next_reusable;

    vy_thread_entry_fn *entry;
    void *param;

    bool needs_join;
};

    #define MAX_THREADS 256
static vy_thread _threads[MAX_THREADS];
static ptrdiff_t _first_reusable = MAX_THREADS;
static ptrdiff_t _next           = 0;

static pthread_mutex_t _guard = PTHREAD_MUTEX_INITIALIZER;

static void *linuxThreadWrapper(void *arg) {
    vy_thread *user_thread  = arg;
    user_thread->needs_join = false;
    user_thread->entry(user_thread->param);
    user_thread->needs_join = true;
    return NULL;
}

VY_DLLEXPORT vy_thread *vySpawnThread(vy_thread_entry_fn *entry, void *param) {
    vy_thread *thrd = NULL;
    pthread_mutex_lock(&_guard);
    if (_first_reusable < MAX_THREADS) {
        thrd            = &_threads[_first_reusable];
        _first_reusable = thrd->next_reusable;
        if (thrd->needs_join) {
            pthread_join(thrd->handle, NULL);
            thrd->needs_join = false;
        }
    } else if (_next < MAX_THREADS) {
        thrd                = &_threads[_next];
        thrd->next_reusable = MAX_THREADS;
        ++_next;
    }
    if (thrd) {
        thrd->entry = entry;
        thrd->param = param;
        if (pthread_create(&thrd->handle, NULL, linuxThreadWrapper, thrd) !=
            0) {
            vyLog("core", "Mutex creation failed");
            thrd = NULL;
        }
    } else {
        vyReportError("core", "Ran out of thread objects");
    }
    pthread_mutex_unlock(&_guard);
    return thrd;
}

VY_DLLEXPORT void vyJoinThread(vy_thread *thread) {
    pthread_join(thread->handle, NULL);
    thread->needs_join = false;
    ptrdiff_t index    = thread - &_threads[0];
    pthread_mutex_lock(&_guard);
    thread->next_reusable = _first_reusable;
    _first_reusable       = index;
    pthread_mutex_unlock(&_guard);
}

#endif
